#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>

void print_file(unsigned char *start, int length) {
  int i;

  printf("Printing file:\n\n");
  for (i=0; i<length; i++) {
    printf("%c", start[i]);
  }
  printf("\n");
}

void modify_file(unsigned char *start, int length) {
  printf("Modifying file..\n\n");
  if (length >= 4) {
    start[0] = 'H';
    start[1] = 'i';
    start[2] = '!';
    start[3] = '\n';
  } else {
    printf("file not big enough to say hello.\n");
    exit(-1);
  }
  printf("...done with modification!\n\n");
}

void sync_file(unsigned char *start, int length) {
  printf("Issuing an msync...\n");
  if (msync(start, length, MS_SYNC|MS_INVALIDATE) != 0) {
    perror("msync failed with: ");
    exit(-1);
  }
  printf("...done with msync!\n");
}

void wait_for_input(void) {
  getchar();
}


int main(int argc, char **argv) {
  int fd;
  struct stat sb;
  unsigned char *mmap_segment;

  // verify somebody gave us a filename as an argument
  if (argc != 2) {
    printf("usage:  mmap_example filename\n");
    return 01;
  }

  // get a file descriptor for "mmap_example.c"
  fd = open(argv[1], O_RDWR);
  if (fd == -1) {
    perror("open failed: ");
    return -1;
  }

  // get attributes of file
  if (fstat(fd, &sb) == -1) {
    perror("fstat failed: ");
    return -1;
  }
  if (sb.st_size == 0) {
    printf("can't mmap a zero-length file\n");
    return 0;
  }

  // use mmap to open up the file
  mmap_segment = (unsigned char *)
    mmap(NULL, sb.st_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
  if (mmap_segment == (unsigned char *) MAP_FAILED) {
    perror("mmap failed:");
    return -1;
  }

  // tell the user where mmap'ed file ended up in the process's
  // address space
  printf("mmap mapped the file to virtual address 0x%08x.\n\n",
         (unsigned int) mmap_segment);
  wait_for_input();

  // print out the contents of the file
  print_file(mmap_segment, sb.st_size);
  wait_for_input();

  // modify the contents of the file.  the modification
  // might reflect into the file immediately, or it might
  // be buffered;  mmap makes no promises.
  modify_file(mmap_segment, sb.st_size);
  wait_for_input();

  // sync the mmap'ed region to the backing file;
  // once sync returns, dirty data in memory is guaranteed
  // to be written back into the file.
  sync_file(mmap_segment, sb.st_size);
  wait_for_input();
  
  return 0;
}
